<?php

namespace Drupal\cms_content_sync;

use Drupal\cms_content_sync\Entity\Flow;
use Drupal\cms_content_sync\Entity\Pool;
use Drupal\cms_content_sync\Event\BeforeEntityTypeExport;
use Drupal\cms_content_sync\Plugin\rest\resource\EntityResource;
use Drupal\cms_content_sync\SyncCore\Storage\ApiStorage;
use Drupal\cms_content_sync\SyncCore\Storage\ConnectionStorage;
use Drupal\cms_content_sync\SyncCore\Storage\ConnectionSynchronizationStorage;
use Drupal\cms_content_sync\SyncCore\Storage\EntityTypeStorage;
use Drupal\cms_content_sync\SyncCore\Storage\InstanceStorage;
use Drupal\cms_content_sync\SyncCore\Storage\PreviewEntityStorage;
use Drupal\encrypt\Entity\EncryptionProfile;
use Drupal\user\Entity\User;
use Drupal\cms_content_sync\Form\PoolForm;

/**
 * Class SyncCoreFlowExport used to export the Synchronization config to the Sync
 * Core backend.
 */
class SyncCoreFlowExport extends SyncCoreExport {

  /**
   * @var \Drupal\cms_content_sync\Entity\Flow
   */
  protected $flow;

  /**
   * Sync Core Config constructor.
   *
   * @param \Drupal\cms_content_sync\Entity\Flow $flow
   *   The flow this exporter is used for.
   */
  public function __construct(Flow $flow) {
    parent::__construct();

    $this->flow = $flow;
  }

  /**
   * Get a list of all Sync Core connections as resource URLs.
   *
   * @return string[]
   */
  protected function getConnectionUrls() {
    $pools = Pool::getAll();
    $urls = [];

    foreach ($this->flow->getEntityTypeConfig() as $id => $type) {
      $entity_type_name = $type['entity_type_name'];
      $bundle_name = $type['bundle_name'];
      $version = $type['version'];

      if ($type['handler'] == Flow::HANDLER_IGNORE) {
        continue;
      }

      $entity_type_pools = [];
      foreach ($type['import_pools'] as $pool_id => $state) {
        if (!isset($entity_type_pools[$pool_id])) {
          $entity_type_pools[$pool_id] = [];
        }
        $entity_type_pools[$pool_id]['import'] = $state;
      }
      foreach ($type['export_pools'] as $pool_id => $state) {
        if (!isset($entity_type_pools[$pool_id])) {
          $entity_type_pools[$pool_id] = [];
        }
        $entity_type_pools[$pool_id]['export'] = $state;
      }

      foreach ($entity_type_pools as $pool_id => $definition) {
        $pool = $pools[$pool_id];
        $export = $definition['export'];
        $import = $definition['import'];

        if ($export == Pool::POOL_USAGE_FORBID && $import == Pool::POOL_USAGE_FORBID) {
          continue;
        }

        $url     = $pool->getBackendUrl();
        $api     = $pool->id;
        $site_id = $pool->getSiteId();

        $local_connection_id = ConnectionStorage::getExternalConnectionId($api, $site_id, $entity_type_name, $bundle_name, $version);
        $urls[] = $url . '/api_unify-api_unify-connection-0_1/' . $local_connection_id;
      }
    }

    return $urls;
  }

  /**
   * Get a list of all Sync Core connection synchronizations as resource URLs.
   *
   * @param array $filters
   *   Additional filters for the synchronization URLs to return.
   *   - entity_type_id: Only return synchronizations for this entity type.
   *    - bundle:         Only return synchronizations for this bundle.
   *    - flow_export:    Only return synchronizations with this export setting.
   *    - flow_import:    Only return synchronizations with this import setting.
   *    - export_pools:   Only return synchronizations with this pool export setting.
   *    - import_pools:   Only return synchronizations with this pool import setting.
   *
   * @return string[]
   */
  protected function getConnectionSynchronizationUrls($filters) {
    $pools = Pool::getAll();
    $urls = [];

    foreach ($this->flow->getEntityTypeConfig() as $id => $type) {
      $entity_type_name = $type['entity_type_name'];
      if (!empty($filters['entity_type_id']) && $filters['entity_type_id'] != $entity_type_name) {
        continue;
      }

      $bundle_name = $type['bundle_name'];
      if (!empty($filters['bundle']) && $filters['bundle'] != $bundle_name) {
        continue;
      }

      $version = $type['version'];

      if ($type['handler'] == Flow::HANDLER_IGNORE) {
        continue;
      }

      if (!empty($filters['flow_export']) && !in_array($type['export'], $filters['flow_export'])) {
        continue;
      }

      if (!empty($filters['flow_import']) && !in_array($type['import'], $filters['flow_import'])) {
        continue;
      }

      $entity_type_pools = [];
      foreach ($type['import_pools'] as $pool_id => $state) {
        if (!isset($entity_type_pools[$pool_id])) {
          $entity_type_pools[$pool_id] = [];
        }
        $entity_type_pools[$pool_id]['import'] = $state;
      }
      foreach ($type['export_pools'] as $pool_id => $state) {
        if (!isset($entity_type_pools[$pool_id])) {
          $entity_type_pools[$pool_id] = [];
        }
        $entity_type_pools[$pool_id]['export'] = $state;
      }

      foreach ($entity_type_pools as $pool_id => $definition) {
        $pool = $pools[$pool_id];
        $export = $definition['export'];
        $import = $definition['import'];

        if (!empty($filters['pool_import']) && !in_array($import, $filters['pool_import'])) {
          continue;
        }

        if (!empty($filters['pool_export']) && !in_array($export, $filters['pool_export'])) {
          continue;
        }

        $url     = $pool->getBackendUrl();
        $api     = $pool->id;
        $site_id = $pool->getSiteId();

        $local_connection_id = ConnectionStorage::getExternalConnectionId($api, $site_id, $entity_type_name, $bundle_name, $version);
        $sync_id = ConnectionSynchronizationStorage::getExternalConnectionSynchronizationId($local_connection_id, !empty($filters['flow_export']) && !empty($filters['pool_export']));
        $urls[] = $url . '/api_unify-api_unify-connection_synchronisation-0_1/' . $sync_id;
      }
    }

    return $urls;
  }

  /**
   * Kindly ask the Sync Core to pull an individual entity, e.g. when the
   * override checkbox is removed.
   *
   * @param string $entity_type_id
   *   The entity type.
   * @param string $bundle
   *   The bundle.
   * @param string $uuid
   *   The entity UUID.
   *
   * @return array
   */
  public function pull($entity_type_id, $bundle, $uuid) {
    $urls = $this->getConnectionSynchronizationUrls(
      [
        'flow_import'     => [ImportIntent::IMPORT_AUTOMATICALLY, ImportIntent::IMPORT_AS_DEPENDENCY],
        'pool_import'     => [Pool::POOL_USAGE_FORCE],
        'entity_type_id'  => $entity_type_id,
        'bundle'          => $bundle,
      ]
    );

    $result = [];

    $type = $this->flow->getEntityTypeConfig($entity_type_id, $bundle);
    $manual = $type['import'] == ImportIntent::IMPORT_MANUALLY;

    foreach ($urls as $url) {
      try {
        $response = $this->client->request(
          'post',
          $url . '/clone/' . $uuid . '?manual=' . ($manual ? 'true' : 'false'),
          ['body' => '']
        );

        if ($response->getStatusCode() != 200) {
          $result[$url] = FALSE;
        }
        elseif (!($body = json_decode($response->getBody(), TRUE))) {
          $result[$url] = FALSE;
        }
        else {
          $result[$url] = $body;
        }
      }
      catch (\Exception $e) {
        $result[$url] = FALSE;
      }
    }

    return $result;
  }

  /**
   * Kindly ask the Sync Core to pull all entities that are auto imported or
   * exported.
   *
   * @param bool $export
   *   Whether to push (TRUE) or pull (FALSE).
   *
   * @return array An associative array with the synchronization URLs as keys.
   *   The value is FALSE for failed requests. Otherwise it's an associative
   *    array with the following keys:
   *    - id:     ID for this process. If none is given, no entities had to be synchronized.
   *    - total:  Total number of entities to be synchronized, if any.
   */
  public function startSync($export) {
    $urls = $this->getConnectionSynchronizationUrls(
      [
        'flow_export' => $export ? [ExportIntent::EXPORT_FORCED] : NULL,
        'pool_export' => $export ? [Pool::POOL_USAGE_FORCE, Pool::POOL_USAGE_ALLOW] : NULL,
        'flow_import' => !$export ? [ImportIntent::IMPORT_AUTOMATICALLY] : NULL,
        'pool_import' => !$export ? [Pool::POOL_USAGE_FORCE] : NULL,
      ]
    );

    $result = [];

    foreach ($urls as $url) {
      try {
        $response = $this->client->request(
          'post',
          $url . '/synchronize?update_all=true',
          ['body' => '']
        );

        if ($response->getStatusCode() != 200) {
          $result[$url] = FALSE;
        }
        elseif (!($body = json_decode($response->getBody(), TRUE))) {
          $result[$url] = FALSE;
        }
        else {
          $result[$url] = $body;
        }
      }
      catch (\Exception $e) {
        $result[$url] = FALSE;
      }
    }

    return $result;
  }

  /**
   * Kindly ask the Sync Core to login to this site.
   */
  public function login() {
    $urls = $this->getConnectionUrls();

    $result = [];

    foreach ($urls as $url) {
      try {
        $response = $this->client->request(
          'post',
          $url . '/login',
          ['body' => '']
        );

        $result[$url] = $response->getStatusCode() == 200 && ($body = json_decode($response->getBody(), TRUE)) && $body['success'];
      }
      catch (\Exception $e) {
        $result[$url] = FALSE;
      }
    }

    return $result;
  }

  /**
   * Create all entity types, connections and synchronizations as required.
   *
   * @throws \Exception If the user profile for import is not available.
   */
  public function prepareBatch() {
    // Ignore disabled flows at export.
    if (!$this->flow->get('status')) {
      return [];
    }

    $export_url = EntityResource::getBaseUrl();
    $enable_preview = static::isPreviewEnabled();

    $cms_content_sync_disable_optimization = boolval(\Drupal::config('cms_content_sync.debug')
      ->get('cms_content_sync_disable_optimization'));

    $user = User::load(CMS_CONTENT_SYNC_USER_ID);
    // During the installation from an existing config for some reason CMS_CONTENT_SYNC_USER_ID is not set right after the installation of the module, so we've to double check that...
    // @ToDo: Why?
    if (is_null(CMS_CONTENT_SYNC_USER_ID)) {
      $user = User::load(\Drupal::service('keyvalue.database')
        ->get('cms_content_sync_user')
        ->get('uid'));
    }

    if (!$user) {
      throw new \Exception(
        t("CMS Content Sync User not found. Encrypted data can't be saved")
      );
    }

    $userData = \Drupal::service('user.data');
    $loginData = $userData->get('cms_content_sync', $user->id(), 'sync_data');

    if (!$loginData) {
      throw new \Exception(t("No credentials for sync user found."));
    }

    $encryption_profile = EncryptionProfile::load(cms_content_sync_PROFILE_NAME);

    foreach ($loginData as $key => $value) {
      $loginData[$key] = \Drupal::service('encryption')
        ->decrypt($value, $encryption_profile);
    }

    $entity_types = $this->flow->sync_entities;

    $pools = Pool::getAll();

    $this->remove(TRUE);

    $operations = [];

    foreach ($this->flow->getEntityTypeConfig() as $id => $type) {
      $entity_type_name = $type['entity_type_name'];
      $bundle_name      = $type['bundle_name'];
      $version          = $type['version'];

      if ($type['handler'] == Flow::HANDLER_IGNORE) {
        continue;
      }
      $handler = $this->flow->getEntityTypeHandler($type);

      $entityFieldManager = \Drupal::service('entity_field.manager');
      /** @var \Drupal\Core\Field\FieldDefinitionInterface[] $fields */
      $fields = $entityFieldManager->getFieldDefinitions($entity_type_name, $bundle_name);

      $entity_type_pools = [];
      foreach ($type['import_pools'] as $pool_id => $state) {
        if (!isset($entity_type_pools[$pool_id])) {
          $entity_type_pools[$pool_id] = [];
        }
        $entity_type_pools[$pool_id]['import'] = $state;
      }
      foreach ($type['export_pools'] as $pool_id => $state) {
        if (!isset($entity_type_pools[$pool_id])) {
          $entity_type_pools[$pool_id] = [];
        }
        $entity_type_pools[$pool_id]['export'] = $state;
      }

      foreach ($entity_type_pools as $pool_id => $definition) {
        $pool   = $pools[$pool_id];
        $export = $definition['export'];
        $import = $definition['import'];

        if ($export == Pool::POOL_USAGE_FORBID && $import == Pool::POOL_USAGE_FORBID) {
          continue;
        }

        $url     = $pool->getBackendUrl();
        $api     = $pool->id;
        $site_id = $pool->getSiteId();

        if (strlen($site_id) > PoolForm::siteIdMaxLength) {
          throw new \Exception(t('The site id of pool ' . $pool_id . ' is having more then ' . PoolForm::siteIdMaxLength . ' characters. This is not allowed due to backend limitations and will result in an exception when it is trying to be exported.'));
        }

        if ($pool->getAuthenticationType() == Pool::AUTHENTICATION_TYPE_BASIC_AUTH) {
          $authentication = [
            'type' => 'basic_auth',
            'username' => $loginData['userName'],
            'password' => $loginData['userPass'],
            'base_url' => $export_url,
          ];
        }
        else {
          $authentication = [
            'type' => 'drupal8_services',
            'username' => $loginData['userName'],
            'password' => $loginData['userPass'],
            'base_url' => $export_url,
          ];
        }

        $entity_type_id = EntityTypeStorage::getExternalEntityTypeId($api, $entity_type_name, $bundle_name, $version);
        $entity_type = [
          'id' => $entity_type_id,
          'name_space' => $entity_type_name,
          'name' => $bundle_name,
          'version' => $version,
          'base_class' => "api-unify/services/drupal/v0.1/models/base.model",
          'custom' => TRUE,
          'new_properties' => [
            'source' => [
              'type' => 'reference',
              'default_value' => NULL,
              'connection_identifiers' => [
                [
                  'properties' => [
                    'id' => 'source_connection_id',
                  ],
                ],
              ],
              'model_identifiers' => [
                [
                  'properties' => [
                    'id' => 'source_id',
                  ],
                ],
              ],
              'multiple' => FALSE,
            ],
            'source_id' => [
              'type' => 'id',
              'default_value' => NULL,
            ],
            'source_connection_id' => [
              'type' => 'id',
              'default_value' => NULL,
            ],
            'preview' => [
              'type' => 'string',
              'default_value' => NULL,
            ],
            'url' => [
              'type' => 'string',
              'default_value' => NULL,
            ],
            'apiu_translation' => [
              'type' => 'object',
              'default_value' => NULL,
            ],
            'metadata' => [
              'type' => 'object',
              'default_value' => NULL,
            ],
            'embed_entities' => [
              'type' => 'object',
              'default_value' => NULL,
              'multiple' => TRUE,
            ],
            'menu_items' => [
              'type' => 'object',
              'default_value' => NULL,
              'multiple' => TRUE,
            ],
            'title' => [
              'type' => 'string',
              'default_value' => NULL,
            ],
            'created' => [
              'type' => 'int',
              'default_value' => NULL,
            ],
            'changed' => [
              'type' => 'int',
              'default_value' => NULL,
            ],
            'uuid' => [
              'type' => 'string',
              'default_value' => NULL,
            ],
          ],
          'new_property_lists' => [
            'list' => [
              '_resource_url' => 'value',
              '_resource_connection_id' => 'value',
              'id' => 'value',
            ],
            'reference' => [
              '_resource_url' => 'value',
              '_resource_connection_id' => 'value',
              'id' => 'value',
            ],
            'details' => [
              '_resource_url' => 'value',
              '_resource_connection_id' => 'value',
              'id' => 'value',
              'source' => 'reference',
              'apiu_translation' => 'value',
              'metadata' => 'value',
              'embed_entities' => 'value',
              'title' => 'value',
              'created' => 'value',
              'changed' => 'value',
              'uuid' => 'value',
              'url' => 'value',
              'menu_items' => 'value',
            ],
            'database' => [
              'id' => 'value',
              'source_id' => 'value',
              'source_connection_id' => 'value',
              'preview' => 'value',
              'url' => 'value',
              'apiu_translation' => 'value',
              'metadata' => 'value',
              'embed_entities' => 'value',
              'title' => 'value',
              'created' => 'value',
              'changed' => 'value',
              'uuid' => 'value',
              'menu_items' => 'value',
            ],
            'modifiable' => [
              'title' => 'value',
              'preview' => 'value',
              'url' => 'value',
              'apiu_translation' => 'value',
              'metadata' => 'value',
              'embed_entities' => 'value',
              'menu_items' => 'value',
            ],
            'required' => [
              'uuid' => 'value',
            ],
          ],
          'api_id' => $api . '-' . ApiStorage::CUSTOM_API_VERSION,
        ];

        $forbidden = $handler->getForbiddenFields();

        foreach ($fields as $key => $field) {
          if (!isset($entity_types[$id . '-' . $key])) {
            continue;
          }

          if (in_array($key, $forbidden)) {
            continue;
          }

          if (isset($entity_type['new_properties'][$key])) {
            continue;
          }

          $entity_type['new_properties'][$key] = [
            'type' => 'object',
            'default_value' => NULL,
            'multiple' => TRUE,
          ];

          $entity_type['new_property_lists']['details'][$key] = 'value';
          $entity_type['new_property_lists']['database'][$key] = 'value';

          $entityFieldManager = \Drupal::service('entity_field.manager');
          $field_definition = $entityFieldManager->getFieldDefinitions($entity_type_name, $bundle_name)[$key];

          if ($field_definition->isRequired()) {
            $entity_type['new_property_lists']['required'][$key] = 'value';
          }

          if (!$field_definition->isReadOnly() || $key == 'menu_link' || $field->getType() == 'path') {
            $entity_type['new_property_lists']['modifiable'][$key] = 'value';
          }
        }

        // TODO entity types should also contain the entity type handler in their machine name, preventing the following potential errors:
        // - Different flows may use different entity type handlers, resulting in different entity type definitions for the same entity type
        // - Changing the entity type handler must change the entity type definition which will not work if we don't update the machine name.
        // - Alternatively, force the same handler to be used at all sites.
        $handler->updateEntityTypeDefinition($entity_type);

        // Dispatch EntityTypeExport event to give other modules the possibility
        // to adjust the entity type definition and add custom fields.
        \Drupal::service('event_dispatcher')->dispatch(BeforeEntityTypeExport::EVENT_NAME, new BeforeEntityTypeExport($entity_type_name, $bundle_name, $entity_type));

        // Create the entity type.
        $operations[] = [$url . '/api_unify-api_unify-entity_type-0_1', [
          'json' => $entity_type,
        ],
        ];

        $pool_connection_id = ConnectionStorage::getExternalConnectionId($api, InstanceStorage::POOL_SITE_ID, $entity_type_name, $bundle_name, $version);
        // Create the pool connection entity for this entity type.
        $operations[] = [$url . '/api_unify-api_unify-connection-0_1', [
          'json' => [
            'id' => $pool_connection_id,
            'name' => 'Drupal pool connection for ' . $entity_type_name . '-' . $bundle_name . '-' . $version,
            'hash' => ConnectionStorage::getExternalConnectionPath($api, InstanceStorage::POOL_SITE_ID, $entity_type_name, $bundle_name, $version),
            'usage' => 'EXTERNAL',
            'status' => 'READY',
            'options' => [
              'update_all' => $cms_content_sync_disable_optimization,
            ],
            'entity_type_id' => $entity_type_id,
          ],
        ],
        ];

        // Create a synchronization from the pool to the preview connection.
        if ($enable_preview) {
          $operations[] = [$url . '/api_unify-api_unify-connection_synchronisation-0_1', [
            'json' => [
              'id' => $pool_connection_id . '--to--preview',
              'name' => 'Synchronization Pool ' . $entity_type_name . '-' . $bundle_name . ' -> Preview',
              'options' => [
                'create_entities' => TRUE,
                'update_entities' => TRUE,
                'delete_entities' => TRUE,
                'update_none_when_loading' => TRUE,
                'exclude_reference_properties' => [
                  'pSource',
                ],
              ],
              'status' => 'READY',
              'source_connection_id' => $pool_connection_id,
              'destination_connection_id' => PreviewEntityStorage::PREVIEW_CONNECTION_ID,
            ],
          ],
          ];
        }

        $crud_operations = [
          'create_item' => [
            'url' => EntityResource::getInternalUrl($api, $entity_type_name, $bundle_name, $version),
          ],
          'update_item' => [
            'url' => EntityResource::getInternalUrl($api, $entity_type_name, $bundle_name, $version, '[id]'),
          ],
          'delete_item' => [
            'url' => EntityResource::getInternalUrl($api, $entity_type_name, $bundle_name, $version, '[id]'),
          ],
        ];
        $connection_options = [
          'authentication' => $authentication,
          'update_all' => $cms_content_sync_disable_optimization,
          'crud' => $crud_operations,
        ];

        if ($export != Pool::POOL_USAGE_FORBID && $type['export'] == ExportIntent::EXPORT_AUTOMATICALLY) {
          $crud_operations['read_list']['url'] = EntityResource::getInternalUrl($api, $entity_type_name, $bundle_name, $version, EntityResource::READ_LIST_ENTITY_ID);
        }

        $local_connection_id = ConnectionStorage::getExternalConnectionId($api, $site_id, $entity_type_name, $bundle_name, $version);
        // Create the instance connection entity for this entity type.
        $operations[] = [$url . '/api_unify-api_unify-connection-0_1', [
          'json' => [
            'id' => $local_connection_id,
            'name' => 'Drupal connection on ' . $site_id . ' for ' . $entity_type_name . '-' . $bundle_name . '-' . $version,
            'hash' => ConnectionStorage::getExternalConnectionPath($api, $site_id, $entity_type_name, $bundle_name, $version),
            'usage' => 'EXTERNAL',
            'status' => 'READY',
            'entity_type_id' => $entity_type_id,
            'instance_id' => $site_id,
            'options' => $connection_options,
          ],
        ],
        ];

        // Create a synchronization from the pool to the local connection.
        if ($import != Pool::POOL_USAGE_FORBID && $type['import'] != ImportIntent::IMPORT_DISABLED) {
          $operations[] = [$url . '/api_unify-api_unify-connection_synchronisation-0_1', [
            'json' => [
              'id' => ConnectionSynchronizationStorage::getExternalConnectionSynchronizationId($local_connection_id, FALSE),
              'name' => 'Synchronization for ' . $entity_type_name . '/' . $bundle_name . '/' . $version . ' from Pool -> ' . $site_id,
              'options' => [
                'dependency_connection_id' => ConnectionStorage::DEPENDENCY_CONNECTION_ID,
                'create_entities' => $type['import'] != ImportIntent::IMPORT_MANUALLY,
                'force_updates' => $cms_content_sync_disable_optimization,
                'update_entities' => TRUE,
                'delete_entities' => boolval($type['import_deletion_settings']['import_deletion']),
                'dependent_entities_only' => $type['import'] == ImportIntent::IMPORT_AS_DEPENDENCY,
                'update_none_when_loading' => TRUE,
                'exclude_reference_properties' => [
                  'pSource',
                ],
              ],
              'status' => 'READY',
              'source_connection_id' => $pool_connection_id,
              'destination_connection_id' => $local_connection_id,
            ],
          ],
          ];
        }
        if ($export != Pool::POOL_USAGE_FORBID && $type['export'] != ExportIntent::EXPORT_DISABLED) {
          $operations[] = [$url . '/api_unify-api_unify-connection_synchronisation-0_1', [
            'json' => [
              'id' => ConnectionSynchronizationStorage::getExternalConnectionSynchronizationId($local_connection_id, TRUE),
              'name' => 'Synchronization for ' . $entity_type_name . '/' . $bundle_name . '/' . $version . ' from ' . $site_id . ' -> Pool',
              'options' => [
                'dependency_connection_id' => ConnectionStorage::POOL_DEPENDENCY_CONNECTION_ID,
                // As entities will only be sent to Sync Core if the sync config
                // allows it, the synchronization entity doesn't need to filter
                // any further
                // 'create_entities' => TRUE,
                // 'update_entities' => TRUE,
                // 'delete_entities' => TRUE,
                // 'dependent_entities_only'  => FALSE,.
                'create_entities' => TRUE,
                'update_entities' => TRUE,
                'delete_entities' => boolval($type['export_deletion_settings']['export_deletion']),
                'force_updates' => $cms_content_sync_disable_optimization,
                'dependent_entities_only' => $export != Pool::POOL_USAGE_FORBID && $type['export'] == ExportIntent::EXPORT_AS_DEPENDENCY,
                'update_none_when_loading' => TRUE,
                'exclude_reference_properties' => [
                  'pSource',
                ],
              ],
              'status' => 'READY',
              'source_connection_id' => $local_connection_id,
              'destination_connection_id' => $pool_connection_id,
            ],
          ],
          ];
        }
      }
    }

    return $operations;
  }

  /**
   * Delete the synchronizations from this connection.
   */
  public function remove($removedOnly = TRUE) {
    return TRUE;

    // @TODO Refactor for pool changes
    $condition   = [
      'operator'  => '==',
      'values'    => [
        [
          'source'  => 'entity',
          'field'   => 'instance_id',
        ],
        [
          'source'  => 'value',
          'value'   => $this->flow->site_id,
        ],
      ],
    ];
    $url         = $this->generateUrl(
      $this->flow->url . '/api_unify-api_unify-connection-0_1',
      [
        'items_per_page'  => '99999',
        'condition' => json_encode($condition),
      ]
    );
    $response    = $this->client->{'get'}($url);
    $body        = json_decode($response->getBody(), TRUE);
    $connections = [];
    foreach ($body['items'] as $reference) {
      $connections[] = $reference['id'];
    }
    $importConnections = $connections;
    $exportConnections = $connections;

    if ($removedOnly) {
      $existingExport = [];
      $existingImport = [];
      foreach ($this->flow->getEntityTypeConfig() as $config) {
        $id = ConnectionStorage::getExternalConnectionId(
          $this->flow->api,
          $this->flow->site_id,
          $config['entity_type_name'],
          $config['bundle_name'],
          $config['version']
        );
        if ($config['export'] != ExportIntent::EXPORT_DISABLED) {
          $existingExport[] = $id;
        }
        if ($config['import'] != ImportIntent::IMPORT_DISABLED) {
          $existingImport[] = $id;
        }
      }
      $importConnections = array_diff($importConnections, $existingImport);
      $exportConnections = array_diff($exportConnections, $existingExport);
    }
    $condition = NULL;
    if (count($exportConnections) > 0) {
      $condition = [
        'operator'    => 'in',
        'values'      => [
          [
            'source'    => 'entity',
            'field'     => 'source_connection_id',
          ],
          [
            'source'    => 'value',
            'value'     => $exportConnections,
          ],
        ],
      ];
    }
    if (count($importConnections) > 0) {
      $importCondition = [
        'operator'    => 'in',
        'values'      => [
          [
            'source'    => 'entity',
            'field'     => 'destination_connection_id',
          ],
          [
            'source'    => 'value',
            'value'     => $importConnections,
          ],
        ],
      ];
      if ($condition) {
        $condition = [
          'operator'    => 'or',
          'conditions'  => [
            $condition,
            $importCondition,
          ],
        ];
      }
      else {
        $condition = $importCondition;
      }
    }

    if (!$condition) {
      return;
    }

    $url = $this->generateUrl(
      $this->flow->url . '/api_unify-api_unify-connection_synchronisation-0_1',
      [
        'condition' => json_encode($condition),
      ]
    );
    $this->client->{'delete'}($url);
  }

}
